const
    ProgramVersion='v0.5.1 beta';
    DefaultPatchScript='PATCH.PAT';
    MaxPatchByteArraySize=256;

type
    TOptions=record
        ScriptFile:Text;
        ScriptBaseDir:String[79];
        Base,Code,Stack:Pointer;
    end;

var
    CurrentVersion:PVersion;

{$IFDEF KEEN1_ARCH}
const
    LevelFilePrefix:PChar='LEVEL';
    LevelDir:array[0..79] of Char=#0;

var
    TargetFile:array[0..79] of Char;

function RedirectFile(FileName:PChar):PChar; assembler;
asm
    PUSH FileName.Word[2]
    PUSH FileName.Word[0]
    PUSH LevelFilePrefix.Word[2]
    PUSH LevelFilePrefix.Word[0]
    MOV AX,5
    PUSH AX
    CALL StrLIComp
    CMP AX,0
    JNE @NoMatch

@Match:
    MOV AX,SEG TargetFile; PUSH AX
    MOV AX,OFFSET TargetFile; PUSH AX
    MOV AX,SEG LevelDir; PUSH AX
    MOV AX,OFFSET LevelDir; PUSH AX
    CALL StrECopy

    PUSH DX
    PUSH AX
    PUSH FileName.Word[2]
    PUSH FileName.Word[0]
    CALL StrECopy

    MOV AX,OFFSET TargetFile
    MOV DX,SEG TargetFile
    JMP @Done

@NoMatch:
    MOV AX,FileName.Word[0]
    MOV DX,FileName.Word[2]

@Done:
end;
{$ENDIF}

{$IFDEF KEEN4_ARCH}
const
    FilePrefix:array[Low(TDataFile)..High(TDataFile)] of PChar=('GAMEMAPS','EGAGRAPH','AUDIO');
    FilePrefixLength:array[Low(TDataFile)..High(TDataFile)] of Word=(8,8,5);
    FileReplacement:array[Low(TDataFile)..High(TDataFile),0..79] of Char=
        ('GAMEMAPS.'+Ext+#0,'EGAGRAPH.'+Ext+#0,'AUDIO.'+Ext+#0);

function RedirectFile(FileName:PChar):PChar;
var
    I:TDataFile;
begin
    RedirectFile:=FileName;
    for I:=Low(TDataFile) to High(TDataFile) do
    begin
        if StrLIComp(FileName,FilePrefix[I],FilePrefixLength[I])=0 then
        begin
            RedirectFile:=@FileReplacement[I];
            Break;
        end;
    end;
end;
{$ENDIF}

procedure Int21HandlerInit; external;
procedure Int21HandlerRun; external;
procedure Int21HandlerDone; external;
{$L REDIR.OBJ}

procedure InitOptions(var Options:TOptions);
var
    ScriptFileName:String[79];
    ExecutableFileName:String[79];
    ExecutableImageLength:Longint;
    CommandLine:String[127];
    I:Integer;

    procedure PrintHelpMessage;
    begin
        Writeln('Commander Keen Patch-Loading Utility '+ProgramVersion);
        Writeln('Copyright (C) Bryan Rodgers <rodgersb@ses.curtin.edu.au>');
        Writeln;
        Writeln('Usage: '+ProgramName+' {patchfile|patchdir} [game arguments]');
    end;
begin
    { Get initial script filespec and command line arguments for game }
    if ParamCount>=1 then
    begin
        ScriptFileName:=ParamStr(1);
        CommandLine:=ParamStrRange(2,ParamCount);
    end else begin
        PrintHelpMessage;
        Halt;
    end;

    { Perform expansion on the script filespec }
    case GetFileType(ScriptFileName) of
        ftNotFound: begin
            Writeln('Error: Can''t find script file ',ScriptFileName);
            Halt;
        end;
        ftFile: ScriptFileName:=FExpand(ScriptFileName);
        ftDirectory: begin
            ScriptFileName:=AdjustRelativePath(ScriptFileName,DefaultPatchScript);
            if GetFileType(ScriptFileName)<>ftFile then
            begin
                Writeln('Error: Can''t find script file ',ScriptFileName);
                Halt;
            end;
        end;
    end;
    Options.ScriptBaseDir:=GetPathOf(ScriptFileName);

    { Attempt to open script file }
    if not ResetWithErrorCheck(Options.ScriptFile,ScriptFileName) then
    begin
        Writeln('Error: Can''t open script file ',ScriptFileName);
        Halt;
    end;

    { Probe executable for version }
    ExecutableFileName:=LocateExecutable(ExecutableFileSpec);
    if LoaderStatus=lsOk then
    begin
        ExecutableImageLength:=GetLoadImageSize(ExecutableFileName);
        if LoaderStatus=lsOk then
        begin
            CurrentVersion:=nil;
            for I:=Low(Version) to High(Version) do
                if Version[I].ImageLength=ExecutableImageLength then
                    CurrentVersion:=@Version[I];
            if CurrentVersion=nil then
            begin
                Writeln('Error: Unrecognised version of ',ExecutableFileName);
                Writeln('(Did you remember to UNLZEXE the executable?)');
                Halt;
            end;
        end else begin
            case LoaderStatus of
                lsFileNotEXE: Writeln('Error: ',ExecutableFileName,' is not a valid EXE file');
                lsDiskError: Writeln('Error: Disk error encountered while loading ',ExecutableFileName);
            end;
            Halt;
        end;
    end else begin
        Writeln('Error: Game executable matching ',ExecutableFileSpec,' not found');
        Halt;
    end;

    Close(Options.ScriptFile);

    { Attempt to load executable into memory }
    LoadExecutable(ExecutableFileName,CommandLine,Options.Base,Options.Code,Options.Stack);
    if LoaderStatus<>lsOk then
    begin
        case LoaderStatus of
            lsOutOfMemory: Writeln('Error: Insufficient memory to load ',ExecutableFileName);
            lsUnknown: Writeln('Error: Unknown error code encountered while loading ',ExecutableFileName);
        end;
        Halt;
    end;

    { Set patching base location in memory }
    SetPatcherBase(Options.Base);

    Reset(Options.ScriptFile);
end;

procedure ParsePatchScript(var Options:TOptions);
var
    Command:String[16];
    VersionEnabled:Boolean;

    procedure PrintErrorHeader;
    begin
        Write('Error: line ',CurrentLineNumber,', command ',Command,': ');
    end;

    procedure TerminateSafely;
    begin
        UnloadExecutable(Options.Base);
        Halt;
    end;

    procedure PrintParserError;
    begin
        PrintErrorHeader;
        case ParserStatus of
            prsEndOfFile,prsMissingParameter: Writeln('Missing expected parameter');
            prsInvalidNumericFormat: Writeln('Invalid numeric format');
            prsParameterTooLong: Writeln('Parameter is too long');
        end;
        TerminateSafely;
    end;

    procedure PrintPatcherError;
    begin
        PrintErrorHeader;
        case PatcherStatus of
            ptsFileNotFound: Writeln('File not found');
            ptsFileCreateError: Writeln('File creation error');
            ptsDiskError: Writeln('Disk error while reading file');
            ptsOutOfSpace: Writeln('Out of disk space');
            ptsMessageTruncated: Writeln('Message is too long');
            ptsUnknown: Writeln('Unknown error');
        end;
        TerminateSafely;
    end;

    procedure PrintLevelError(Level:Integer);
    begin
        PrintErrorHeader;
        Writeln('No string exists for level ',Level);
        TerminateSafely;
    end;

    procedure PatchInline;
    var
        Offset:Longint;
        Data:array[0..MaxPatchByteArraySize] of Byte;
        Length:Word;
    begin
        Offset:=GetLongint(Options.ScriptFile);
        if ParserStatus=prsOk then
        begin
            GetByteArray(Options.ScriptFile,Length,Data);
            if ParserStatus=prsOk then
            begin
                if VersionEnabled then PatchBytes(Offset,Length,Data);
            end else PrintParserError;
        end else PrintParserError;
    end;

    procedure PatchFromFile;
    var
        Offset:Longint;
        InputFileName:String[79];
    begin
        Offset:=GetLongint(Options.ScriptFile);
        if ParserStatus=prsOk then
        begin
            InputFileName:=GetWord(Options.ScriptFile);
            if ParserStatus=prsOk then
            begin
                if VersionEnabled then
                begin
                    InputFileName:=AdjustRelativePath(Options.ScriptBaseDir,InputFileName);
                    PatchFile(Offset,InputFileName);
                    if PatcherStatus<>ptsOk then PrintPatcherError;
                end;
            end else PrintParserError;
        end else PrintParserError;
    end;

    procedure GetVersion;
    var
        NewVersion:String[3];
    begin
        NewVersion:=LowerCase(GetWord(Options.ScriptFile));
        if ParserStatus=prsOk then
        begin
            VersionEnabled:=(NewVersion=CurrentVersion^.ID) or (NewVersion='all');
        end else PrintParserError;
    end;

    procedure DumpImage;
    var
        OutputFileName:String[79];
        Offset,Length:Longint;
    begin
        OutputFileName:=GetWord(Options.ScriptFile);
        if ParserStatus=prsOk then
        begin
            OutputFileName:=AdjustRelativePath(Options.ScriptBaseDir,OutputFileName);
            Offset:=GetLongint(Options.ScriptFile);
            if ParserStatus=prsOk then
            begin
                Length:=GetLongint(Options.ScriptFile);
                if ParserStatus<>prsOk then Length:=CurrentVersion^.ImageLength-Offset;
            end else begin
                Offset:=0;
                Length:=CurrentVersion^.ImageLength;
            end;
            if VersionEnabled then
            begin
                Writeln('Writing program image from offset ',Offset,
                    ', length ',Length,' to file ',OutputFileName);
                WriteImage(Offset,Length,OutputFileName);
                if PatcherStatus<>ptsOk then PrintPatcherError;
            end;
        end else PrintParserError;
    end;

{$IFDEF KEEN1_ARCH}
    procedure PatchLevelDir;
    var
        NewDir:String[79];
    begin
        NewDir:=GetWord(Options.ScriptFile);
        if ParserStatus=prsOk then
        begin
            if VersionEnabled then
            begin
                StrPCopy(LevelDir,DropTrailingBackslash(AdjustRelativePath(Options.ScriptBaseDir,NewDir))+'\');
            end;
        end else PrintParserError;
    end;
{$ENDIF}

{$IFDEF KEEN1}
    procedure PatchLevelHint;
    var
        Level:Integer;
        Message:String;
    begin
        Level:=GetLongint(Options.ScriptFile);
        if ParserStatus=prsOk then
        begin
            Message:=GetText(Options.ScriptFile);
            if (Level>=1) and (Level<=16) and (CurrentVersion^.HintTextGridLineIndex^[Level]<>$FF) then
            begin
                if VersionEnabled then
                begin
                    PatchTextGrid(CurrentVersion^.HintOffset,
                        CurrentVersion^.HintTextGridLineIndex^[Level],
                        CurrentVersion^.HintTextGridLineSpan^[Level],
                        CurrentVersion^.HintTextGridLineSize^,
                        Message);
                    if PatcherStatus<>ptsOk then PrintPatcherError;
                end;
            end else PrintLevelError(Level);
        end else PrintPatcherError;
    end;
{$ENDIF}

{$IFDEF KEEN4_ARCH}
    procedure PatchLevelString(LevelStringOffset:Longint;LevelStringLength:PLevelStringLength);
    var
        Level,I:Integer;
        Message:String;
        Offset:Longint;
    begin
        Level:=GetLongint(Options.ScriptFile);
        if ParserStatus=prsOk then
        begin
            Message:=GetText(Options.ScriptFile);
            if (Level>=0) and (Level<=NumLevels) then
            begin
                if VersionEnabled then
                begin
                    Offset:=LevelStringOffset;
                    for I:=0 to Level-1 do Inc(Offset,LevelStringLength^[I]);
                    PatchASCIIZ(Offset,LevelStringLength^[Level],Message);
                    if PatcherStatus<>ptsOk then PrintPatcherError;
                end;
            end else PrintLevelError(Level);
        end else PrintParserError;
    end;

    procedure PatchDataFile(DataFile:TDataFile);
    var
        NewFileName:String[79];
    begin
        NewFileName:=GetWord(Options.ScriptFile);
        if ParserStatus=prsOk then
        begin
            if VersionEnabled then
            begin
                StrPCopy(FileReplacement[DataFile],AdjustRelativePath(Options.ScriptBaseDir,NewFileName));
            end;
        end else PrintParserError;
    end;

    procedure PatchDataResource(DataResource:TDataResource);
    var
        InputFileName:String[79];
    begin
        InputFileName:=GetWord(Options.ScriptFile);
        if ParserStatus=prsOk then
        begin
            if VersionEnabled then
            begin
                InputFileName:=AdjustRelativePath(Options.ScriptBaseDir,InputFileName);
                if CurrentVersion^.DataResourceOffset^[DataResource]<>$FFFFFFFF then
                begin
                    PatchFile(CurrentVersion^.DataResourceOffset^[DataResource],InputFileName);
                    if PatcherStatus<>ptsOk then PrintPatcherError;
                end;
            end;
        end else PrintParserError;
    end;
{$ENDIF}

{$IFDEF KEEN4}
    procedure PatchLevelHint;
    var
        Level,Hint:Integer;
        Message:String;
    begin
        Level:=GetLongint(Options.ScriptFile);
        if ParserStatus=prsOk then
        begin
            Message:=GetText(Options.ScriptFile);
            Hint:=LevelHintMap[Level];
            if Hint<>$FF then
            begin
                if VersionEnabled then
                begin
                    PatchASCIIZ(CurrentVersion^.LevelHintOffset^[Hint],
                        CurrentVersion^.LevelHintLength^[Hint],
                        Message);
                end;
            end else PrintLevelError(Level);
        end else PrintPatcherError;
    end;
{$ENDIF}

begin
    VersionEnabled:=True;

    repeat
        Command:=GetCommand(Options.ScriptFile);
        if ParserStatus=prsOK then
        begin
            if Command='%ext' then
            begin
                GetWord(Options.ScriptFile);
                if ParserStatus<>prsOk then PrintParserError;
            end else if Command='%patch' then begin
                PatchInline;
            end else if Command='%patchfile' then begin
                PatchFromFile;
            end else if Command='%version' then begin
                GetVersion;
            end else if Command='%dump' then begin
                DumpImage;
            end else if Command='%abort' then begin
                Writeln('%abort encountered on line ',CurrentLineNumber);
                Writeln('Patching halted at this point.');
                TerminateSafely;
            end else if Command='%end' then begin
                Break;
        {$IFDEF KEEN1_ARCH}
            end else if Command='%level.dir' then begin
                PatchLevelDir;
        {$ENDIF}
        {$IFDEF KEEN1}
            end else if Command='%level.hint' then begin
                PatchLevelHint;
        {$ENDIF}
        {$IFDEF KEEN4_ARCH}
            end else if Command='%level.name' then begin
                PatchLevelString(CurrentVersion^.LevelNameOffset,CurrentVersion^.LevelNameLength);
            end else if Command='%level.entry' then begin
                PatchLevelString(CurrentVersion^.LevelEntryOffset,CurrentVersion^.LevelEntryLength);
            end else if Command='%gamemaps' then begin
                PatchDataFile(dfGameMaps);
            end else if Command='%egagraph' then begin
                PatchDataFile(dfEGAGraph);
            end else if Command='%audio' then begin
                PatchDataFile(dfAudio);
            end else if Command='%audiohed' then begin
                PatchDataResource(drAudioHed);
            end else if Command='%egahead' then begin
                PatchDataResource(drEGAHead);
            end else if Command='%maphead' then begin
                PatchDataResource(drMapHead);
            end else if Command='%audiodct' then begin
                PatchDataResource(drAudioDct);
            end else if Command='%egadict' then begin
                PatchDataResource(drEGADict);
        {$ENDIF}
        {$IFDEF KEEN4}
            end else if Command='%level.hint' then begin
                PatchLevelHint;
        {$ENDIF}
            end else begin
                PrintErrorHeader;
                Writeln('Unrecognised command ''',Command,'''');
                TerminateSafely;
            end;
        end else Break;
    until False;

    Close(Options.ScriptFile);
end;

procedure RunGame(var Options:TOptions);
begin
    Int21HandlerRun;
    RunExecutable(Options.Base,Options.Code,Options.Stack);
end;

procedure Main;
var
    Options:TOptions;
begin
    SwapVectors;
    Int21HandlerInit;
    InitOptions(Options);
    ParsePatchScript(Options);
    RunGame(Options);
    Int21HandlerDone;
    SwapVectors;
end;
