unit Loader;

{$A+,B+,E-,F-,G-,I-,P+,Q-,R-,S-,T-,V-,X+}

{ Simple unit for loading DOS executables in memory }

interface

type
    TLoaderStatus=(
        lsOk,
        lsFileNotFound,
        lsFileNotEXE,
        lsDiskError,
        lsOutOfMemory,
        lsUnknown
        );

const
{ Error status variable - Most functions will set this on return }
    LoaderStatus:TLoaderStatus=lsOk;

{ Returns first filename to match given wildcard, or blank if not found }
function LocateExecutable(const FileSpec:String):String;

{ Returns load image size for given executable file, 0 if not a valid .EXE }
function GetLoadImageSize(const FileName:String):Longint;

{ Loads executable into memory but does not execute. Returns pointers to    }
{ image start, code entry point and initial stack location.                 }
procedure LoadExecutable(const FileName,CommandLine:String;var Base,Code,Stack:Pointer);

{ Runs loaded executable in memory }
procedure RunExecutable(Base,Code,Stack:Pointer);

{ Unloads executable from memory (only call if it hasn't executed yet) }
procedure UnloadExecutable(Base:Pointer);

implementation

uses Dos,Strings;

const
    ValidEXESignature=$5A4D;

type
{ .EXE file header structure }
    TEXEHeader=record
        ID:Word;
        ImageLengthLow:Word;
        ImageLengthHigh:Word;
        NumRelocs:Word;
        HeaderSize:Word;
        MinParagraphs:Word;
        MaxParagraphs:Word;
        InitSS:Word;
        InitSP:Word;
        Checksum:Word;
        InitIP:Word;
        InitCS:Word;
        FirstRelocOffset:Word;
        OverlayNumber:Word;
    end;

{ Structure required for DOS INT 21H/4BH Exec call }
    TExec=record
        Environment:Word;
        CommandLine:Pointer;
        FCB:array[1..2] of Pointer;
        Stack:Pointer;
        Code:Pointer;
    end;

{ Structure used to access 16-bit components of a doubleword }
    MemLong=record
        Lo,Hi:Word;
    end;

{ Structure used to access 16-bit segment/offset components of a pointer }
    MemPtr=record
        MemOfs,MemSeg:Word;
    end;

var
{ Storage space for stack and base pointers and stack segment }
    SaveSS,SaveSP,SaveBP:Word;

function LocateExecutable(const FileSpec:String):String;
var
    Search:SearchRec;
begin
    FindFirst(FileSpec,AnyFile,Search);
    if DosError=0 then
    begin
        LocateExecutable:=Search.Name;
        LoaderStatus:=lsOk;
    end else begin
        LocateExecutable:='';
        LoaderStatus:=lsFileNotFound;
    end;
end;

function GetLoadImageSize(const FileName:String):Longint;
var
    EXEHeader:TEXEHeader;
    InputFile:File;
begin
    GetLoadImageSize:=0;
    FileMode:=0;
    Assign(InputFile,FileName);
    Reset(InputFile,1);
    if IOResult=0 then
    begin
        BlockRead(InputFile,EXEHeader,SizeOf(TEXEHeader));
        if IOResult=0 then
        begin
            if EXEHeader.ID=ValidEXESignature then
            begin
                GetLoadImageSize:=Longint(EXEHeader.ImageLengthHigh-1)*512+
                    Longint(EXEHeader.ImageLengthLow)-Longint(EXEHeader.HeaderSize*16);
                LoaderStatus:=lsOk;
            end else LoaderStatus:=lsFileNotEXE;
        end else LoaderStatus:=lsDiskError;
        Close(InputFile);
    end else LoaderStatus:=lsFileNotFound;
end;

procedure LoadExecutable(const FileName,CommandLine:String;var Base,Code,Stack:Pointer);
var
    ChildPrefixSeg:Word;
    ASCIIZProgramName:array[0..63] of Char;
    ASCIIZCommandLine:array[0..127] of Char;
    ProgramExec:TExec;
begin
    StrPCopy(ASCIIZProgramName,FileName);
    StrPCopy(ASCIIZCommandLine,CommandLine);
    ProgramExec.Environment:=0;
    ProgramExec.CommandLine:=@ASCIIZCommandLine;
    ProgramExec.FCB[1]:=Ptr(PrefixSeg,$005C);
    ProgramExec.FCB[2]:=Ptr(PrefixSeg,$006C);
    asm
    { Load executable as child process into memory, but do not execute }
        PUSH DS
        MOV AX,SS
        MOV DS,AX
        MOV ES,AX
        LEA DX,ASCIIZProgramName
        LEA BX,ProgramExec
        MOV AX,$4B01
        INT $21
        POP DS
        JNC @NoError
        @Error:
            MOV DOSError,AX
            JMP @Done
        @NoError:
            MOV DOSError,0
        @Done:
    end;

    if DOSError=0 then
    begin
        asm
        { Get the prefix segment of the child process }
            MOV AH,$62
            INT $21
            MOV ChildPrefixSeg,BX

        { Set current prefix segment back to our own }
            MOV BX,PrefixSeg
            MOV AH,$50
            INT $21
        end;
        Base:=Ptr(ChildPrefixSeg+$10,$0000);
        Code:=ProgramExec.Code;
        Stack:=ProgramExec.Stack;
        LoaderStatus:=lsOk;
    end else begin
        case DOSError of
            2: LoaderStatus:=lsFileNotFound;
            8: LoaderStatus:=lsOutOfMemory;
            else LoaderStatus:=lsUnknown;
        end;
    end;
end;

procedure RunExecutable(Base,Code,Stack:Pointer); assembler;
asm
{ Set active process back to child }
    MOV BX,MemPtr(Base).MemSeg
    SUB BX,$10
    MOV AH,$50
    INT $21

{ Set INT 22 return address for child process }
    MOV BX,MemPtr(Base).MemSeg
    SUB BX,$10
    MOV ES,BX
    MOV SI,BX
    MOV MemPtr(ES:[$000A]).MemOfs,OFFSET @ReturnPoint
    MOV MemPtr(ES:[$000A]).MemSeg,SEG @ReturnPoint

{ Preserve SS:SP/SS:BP stack pointer and base pointer }
    MOV SaveBP,BP
    MOV SaveSS,SS
    MOV SaveSP,SP

{ Set up child stack and entry point }
    MOV AX,MemPtr(Stack).MemOfs
    MOV DX,MemPtr(Stack).MemSeg
    MOV CX,MemPtr(Code).MemOfs
    MOV BX,MemPtr(Code).MemSeg
    MOV SS,DX
    MOV SP,AX
    PUSH BX
    PUSH CX

{ Clear registers and jump to entry point }
    MOV DS,SI
    MOV ES,SI
    MOV AX,0
    MOV BX,0
    MOV CX,0
    MOV DX,0
    MOV SI,0
    MOV DI,0
    MOV BP,0
    RETF

{ --- Child program will return here when it terminates --- }
@ReturnPoint:

{ Restore data/stack segments and stack/base pointer }
    MOV AX,SEG @Data
    MOV DS,AX
    MOV BP,SaveBP
    MOV SS,SaveSS
    MOV SP,SaveSP
end;

procedure UnloadExecutable(Base:Pointer); assembler;
asm
    MOV AX,MemPtr(Base).MemSeg
    SUB AX,$10
    MOV ES,AX
    MOV AH,$49
    INT $21
end;

end.
