unit Parser;

{$A+,B+,E-,F-,G-,I-,P+,Q-,R-,S-,T-,V-,X+}

{ Simple patch script parsing unit }

interface

type
    TParserStatus=(
        prsOK,
        prsEndOfFile,
        prsInvalidNumericFormat,
        prsMissingParameter,
        prsParameterTooLong
        );

const
{ Error status variable, set by all procedures here on return }
    ParserStatus:TParserStatus=prsOk;

{ Current line number }
    CurrentLineNumber:Word=0;

{ Scans for and returns a macro command (prefixed by '%'). Ignores any      }
{ junk that may be encountered. String returned includes leading '%' and is }
{ converted to lowercase. File pointer is left at whitespace following      }
{ macro command. Check ParserStatus variable to find out if end of file is  }
{ reached.                                                                  }
function GetCommand(var F:Text):String;

{ Scans for and returns a long integer (can be in decimal format or have a  }
{ leading '$' or '0x' to denote hex notation). Whitespace leading up to     }
{ number is ignored, file pointer is left at whitespace following integer.  }
{ Check ParserStatus variable to find out if the number is valid format.    }
function GetLongint(var F:Text):Longint;

{ Scans for and returns a word string, delimted by whitespace at both ends. }
{ Backslash escapes are only processed if the word is quoted with double    }
{ quote characters. Therefore backslashes in filenames will not be          }
{ interpreted unless in quotes, and any hashes that appear unquoted will    }
{ be interpreted as comments. If the word string is the first word on the   }
{ line and starts with a '%' character, it will be seen as a macro command  }
{ instead and the word will be considered non-existant (blank string        }
{ returned and ParserStatus=prsMissingParameter).                           }
function GetWord(var F:Text):String;

{ Reads in a stream of characters and returns a string. Leading/trailing    }
{ whitespace at start/end of stream and at the end of each line are         }
{ removed. Embedded EOL characters are returned as a newline (#10)          }
{ character embedded inside the string. The stream of characters is         }
{ terminated either by a macro command or the end of file. Backslash        }
{ escapes are always processed. Unescaped hashes are considered comments to }
{ the end of line. Leading '%' characters must be escaped to avoid          }
{ interpretation as a macro command. The text stream may be up to a maximum }
{ of 255 characters.                                                        }
function GetText(var F:Text):String;

{ Reads in a stream of characters and returns a byte array. The stream of   }
{ characters is expected to be whitespace-seperated words, each word is     }
{ either an integer (parseable by GetInteger()) or a string surrounded by   }
{ double quotes which may contain backslash escape sequences. Each integer  }
{ encountered is considered a single byte (0..255) and each character (or   }
{ backslash escape) in a quoted string is considered a singe byte. The      }
{ Length parameter on return it contains number of bytes actually read. The }
{ sequence of bytes is terminated either by a macro command or end of file. }
{ Check the ParserStatus variable for possible numeric format errors.       }
procedure GetByteArray(var F:Text;var Length:Word;var Data:array of Byte);

implementation

const
    TabSize=8;

type
    TReadMode=(
        rmDone,
        rmLookingForCommand,
        rmReadingCommand,
        rmLookingForKeyword,
        rmReadingKeyword,
        rmReadingQuotedString,
        rmLookingForTextStream,
        rmReadingTextStream
        );

const
    ReadMode:TReadMode=rmDone;

function LoCase(C:Char):Char;
begin
    if C in ['A'..'Z'] then LoCase:=Chr(Ord(C)+32) else LoCase:=C;
end;

function GetNextChar(var F:Text):Char;
const
    LastChar:Char=#13;
    ThisChar:Char=#10;
    CurrentColumn:Word=0;
    PendingSpaces:Word=0;
var
    C:Char;

    function ReadChar(var F:Text):Char;
    begin
        if PendingSpaces>0 then
        begin
            LastChar:=ThisChar;
            ThisChar:=#32;
            Dec(PendingSpaces);
        end else begin
            LastChar:=ThisChar;
            repeat
                Read(F,ThisChar);
                case ThisChar of
                    #13:;
                    #9: begin
                        PendingSpaces:=TabSize-CurrentColumn mod TabSize-1;
                        ThisChar:=#32;
                        Break;
                    end;
                    else Break;
                end;
            until False;
        end;

        ReadChar:=ThisChar;
        Inc(CurrentColumn);
        if LastChar=#10 then
        begin
            Inc(CurrentLineNumber);
            CurrentColumn:=1;
        end;
    end;
    procedure SkipToEol(var F:Text);
    begin
        repeat until ReadChar(F) in [#10,#26];
    end;
    function BackslashEscape(C:Char):Char;
    begin
        case C of
            '0': BackslashEscape:=#0;
            'a': BackslashEscape:=#7;
            'b': BackslashEscape:=#8;
            't': BackslashEscape:=#9;
            'n': BackslashEscape:=#10;
            'r': BackslashEscape:=#13;
            else BackslashEscape:=C;
        end;
    end;
begin
    case ReadMode of
        rmLookingForCommand: begin
            repeat
                C:=ReadChar(F);
                case C of
                    '#': SkipToEol(F);
                    '%': begin
                        if LastChar in [#10,#32] then
                        begin
                            ReadMode:=rmReadingCommand;
                            Break;
                        end;
                    end;
                    #26: begin
                        ReadMode:=rmDone;
                        Break;
                    end;
                end;
            until False;
            GetNextChar:=C;
        end;
        rmReadingCommand: begin
            C:=ReadChar(F);
            case C of
                '#': begin
                    SkipToEol(F);
                    ReadMode:=rmDone;
                end;
                #10,#26,#32: ReadMode:=rmDone;
            end;
            GetNextChar:=C;
        end;
        rmLookingForKeyword: begin
            repeat
                C:=ReadChar(F);
                case C of
                    '#': SkipToEol(F);
                    #10,#32:;
                    #26: begin
                        ReadMode:=rmDone;
                        Break;
                    end;
                    '"': begin
                        ReadMode:=rmReadingQuotedString;
                        C:=GetNextChar(F);
                        Break;
                    end;
                    '%': begin
                        if LastChar in [#10,#32] then
                        begin
                            ReadMode:=rmReadingCommand;
                            Break;
                        end;
                    end;
                    else begin
                        ReadMode:=rmReadingKeyword;
                        Break;
                    end;
                end;
            until False;
            GetNextChar:=C;
        end;
        rmReadingKeyword: begin
            C:=ReadChar(F);
            case C of
                '#': begin
                    SkipToEol(F);
                    ReadMode:=rmDone;
                end;
                '"': begin
                    ReadMode:=rmReadingQuotedString;
                    C:=GetNextChar(F);
                end;
                #10,#26,#32: ReadMode:=rmDone;
            end;
            GetNextChar:=C;
        end;
        rmReadingQuotedString: begin
            C:=ReadChar(F);
            case C of
                '"': begin
                    ReadMode:=rmReadingKeyword;
                    C:=GetNextChar(F);
                end;
                '\': C:=BackslashEscape(ReadChar(F));
                #26: ReadMode:=rmDone;
            end;
            GetNextChar:=C;
        end;
        rmLookingForTextStream: begin
            repeat
                C:=ReadChar(F);
                case C of
                    '#': SkipToEol(F);
                    '%': begin
                        if LastChar in [#10,#32] then
                        begin
                            ReadMode:=rmReadingCommand;
                            Break;
                        end;
                    end;
                    #10,#32:;
                    #26: begin
                        ReadMode:=rmDone;
                        Break;
                    end;
                    '\': begin
                        C:=BackslashEscape(ReadChar(F));
                        ReadMode:=rmReadingTextStream;
                        Break;
                    end;
                    else begin
                        ReadMode:=rmReadingTextStream;
                        Break;
                    end;
                end;
            until False;
            GetNextChar:=C;
        end;
        rmReadingTextStream: begin
            C:=ReadChar(F);
            case C of
                '#': begin
                    SkipToEol(F);
                    C:=#10;
                end;
                '%': if LastChar in [#10,#32] then ReadMode:=rmReadingCommand;
                '\': C:=BackslashEscape(ReadChar(F));
                #26: ReadMode:=rmDone;
            end;
            GetNextChar:=C;
        end;
    end;
end;

function GetCommand(var F:Text):String;
var
    S:String;
    C:Char;
begin
    if ReadMode<>rmReadingCommand then
    begin
        ReadMode:=rmLookingForCommand;
        S:=GetNextChar(F);
    end else S:='%';

    if ReadMode=rmReadingCommand then
    begin
        repeat
            C:=LoCase(GetNextChar(F));
            if ReadMode<>rmReadingCommand then Break;
            S:=S+C;
        until False;
        ParserStatus:=prsOk;
    end else ParserStatus:=prsEndOfFile;

    GetCommand:=S;
end;

function GetLongint(var F:Text):Longint;
var
    S:String;
    N:Longint;
    Code:Integer;
begin
    S:=GetWord(F);

    if ParserStatus=prsOk then
    begin
        if Copy(S,1,2)='0x' then
        begin
            Delete(S,1,2);
            Insert('$',S,1);
        end;
        Val(S,N,Code);
        if Code<>0 then
        begin
            ParserStatus:=prsInvalidNumericFormat;
            GetLongint:=0;
        end else GetLongint:=N;
    end else GetLongint:=0;
end;

function GetWord(var F:Text):String;
var
    S:String;
    C:Char;
begin
    S:='';

    if ReadMode<>rmReadingCommand then
    begin
        ReadMode:=rmLookingForKeyword;
        C:=GetNextChar(F);
        while not (ReadMode in [rmDone,rmReadingCommand]) do
        begin
            S:=S+C;
            C:=GetNextChar(F);
        end;

        if Length(S)=0 then
        begin
            case ReadMode of
                rmReadingCommand: ParserStatus:=prsMissingParameter;
                rmDone: ParserStatus:=prsEndOfFile;
            end;
        end else ParserStatus:=prsOk;
    end else ParserStatus:=prsMissingParameter;

    GetWord:=S;
end;

function GetText(var F:Text):String;
var
    S:String;
    C:Char;
    PendingSpaces,PendingNewlines:Word;
begin
    S:='';
    PendingSpaces:=0;
    PendingNewlines:=0;

    if ReadMode<>rmReadingCommand then
    begin
        ReadMode:=rmLookingForTextStream;
        C:=GetNextChar(F);
        ParserStatus:=prsOk;

        if ReadMode=rmReadingTextStream then
        begin
            while ReadMode=rmReadingTextStream do
            begin
                case C of
                    #32: Inc(PendingSpaces);
                    #10: begin
                        PendingSpaces:=0;
                        Inc(PendingNewlines);
                    end;
                    else begin
                        while PendingNewlines>0 do
                        begin
                            S:=S+#10;
                            Dec(PendingNewlines);
                        end;
                        while PendingSpaces>0 do
                        begin
                            S:=S+#32;
                            Dec(PendingSpaces);
                        end;
                        S:=S+C;
                    end;
                end;
                C:=GetNextChar(F);
            end;
        end;
    end else ParserStatus:=prsOK;

    GetText:=S;
end;

procedure GetByteArray(var F:Text;var Length:Word;var Data:array of Byte);
var
    S:String;
    C:Char;
    N:Longint;
    Code:Integer;
begin
    Length:=0;
    ParserStatus:=prsOk;

    while (ReadMode<>rmReadingCommand) and (ParserStatus=prsOk) and not Eof(F) do
    begin
        ReadMode:=rmLookingForKeyword;
        C:=GetNextChar(F);
        case ReadMode of
            rmReadingKeyword: begin
                S:='';
                while ReadMode in [rmReadingKeyword,rmReadingQuotedString] do
                begin
                    S:=S+C;
                    C:=GetNextChar(F);
                end;
                if Copy(S,1,2)='0x' then
                begin
                    Delete(S,1,2);
                    Insert('$',S,1);
                end;
                Val(S,N,Code);
                if Code<>0 then
                begin
                    ParserStatus:=prsInvalidNumericFormat;
                end else if Length<High(Data)-Low(Data)+1 then begin
                    Data[Length]:=N;
                    Inc(Length);
                end else ParserStatus:=prsParameterTooLong;
            end;
            rmReadingQuotedString: begin
                while ReadMode in [rmReadingKeyword,rmReadingQuotedString] do
                begin
                    if Length<High(Data)-Low(Data)+1 then
                    begin
                        Data[Length]:=Ord(C);
                        Inc(Length);
                    end else ParserStatus:=prsParameterTooLong;
                    C:=GetNextChar(F);
                end;
            end;
        end;
    end;

    if Length=0 then
    begin
        if ReadMode=rmReadingCommand then
        begin
            ParserStatus:=prsMissingParameter;
        end else if ParserStatus=prsOk then begin
            ParserStatus:=prsEndOfFile;
        end;
    end;
end;

end.
