unit Patcher;

{$A+,B+,E-,F-,G-,I-,P+,Q-,R-,S-,T-,V-,X+}

{ Simple unit for patching DOS executables in memory }

interface

type
    TPatcherStatus=(
        ptsOk,
        ptsFileNotFound,
        ptsFileCreateError,
        ptsDiskError,
        ptsOutOfSpace,
        ptsMessageTruncated,
        ptsUnknown
        );

const
{ Error status variable, set by all functions appropiately }
    PatcherStatus:TPatcherStatus=ptsOk;

{ Sets base memory location for patching offsets }
procedure SetPatcherBase(NewBase:Pointer);

{ Writes an array of null-terminated space-padded strings of varying length }
procedure PatchTextGrid(Offset:Longint;StartLine,NumLines:Integer;
    var LineSize:array of Byte;const Message:String);

{ Writes a null-terminated string }
procedure PatchASCIIZ(Offset:Longint;MaxLength:Word;const Message:String);

{ Loads patch data from a binary file }
procedure PatchFile(Offset:Longint;const FileName:String);

{ Patches from an array of byte values }
procedure PatchBytes(Offset:Longint;Length:Word;var Data:array of Byte);

{ Dumps program image to a disk file for debugging purposes }
procedure WriteImage(Offset,Length:Longint;const FileName:String);

implementation

uses Strings;

const
{ Maximum size for BlockRead/BlockWrite requests }
    TransferBlockSize=65520;

type
{ Structure used to access 16-bit components of a doubleword }
    MemLong=record
        Lo,Hi:Word;
    end;

{ Structure used to access 16-bit segment/offset components of a pointer }
    MemPtr=record
        MemOfs,MemSeg:Word;
    end;

var
{ Base memory pointer for patching routines }
    PatcherBase:Pointer;

{ Returns a normalised pointer }
function Normalise(P:Pointer):Pointer; assembler;
asm
    MOV BX,MemPtr(P).MemOfs
    MOV DX,MemPtr(P).MemSeg
    MOV AX,BX
    AND AX,$000F
    MOV CL,4
    SHR BX,CL
    ADD DX,BX
end;

{ Returns PatcherBase+Offset as a normalised pointer }
function PointerOffset(LinearOffset:Longint):Pointer; assembler;
asm
    MOV AX,16
    MUL MemPtr(PatcherBase).MemSeg
    ADD AX,MemPtr(PatcherBase).MemOfs
    ADC DX,$0000
    ADD AX,MemLong(LinearOffset).Lo
    ADC DX,MemLong(LinearOffset).Hi
    AND DX,$000F
    MOV BX,AX
    AND BX,$000F
    MOV CX,16
    DIV CX
    MOV DX,AX
    MOV AX,BX
end;

procedure SetPatcherBase(NewBase:Pointer);
begin
    PatcherBase:=NewBase;
end;

procedure PatchTextGrid(Offset:Longint;StartLine,NumLines:Integer;
    var LineSize:array of Byte;const Message:String);
var
    Dest:PChar;
    SourceIndex,DestIndex:Integer;
    CurrentRow,CurrentColumn:Integer;
    I:Integer;
begin
    for I:=0 to StartLine-1 do Inc(Offset,LineSize[I]);
    Dest:=PointerOffset(Offset);
    CurrentColumn:=0;
    CurrentRow:=0;
    SourceIndex:=1;
    DestIndex:=0;
    PatcherStatus:=ptsOk;

    { Each loop deals with one character in Message string }
    while (SourceIndex<=Length(Message)) and (CurrentRow<NumLines) do
    begin
        { Check for newline (#10) character }
        if Message[SourceIndex]=#10 then
        begin
            { Pad current line with spaces, then terminate with #10#0,  }
            { and start onto a new line.                                }
            while CurrentColumn<LineSize[StartLine+CurrentRow]-2 do
            begin
                Dest[DestIndex]:=#32;
                Inc(DestIndex);
                Inc(CurrentColumn);
            end;
            Dest[DestIndex]:=#10;
            Dest[DestIndex+1]:=#0;
            Inc(DestIndex,2);
            CurrentColumn:=0;
            Inc(CurrentRow);
        end else begin
            { Otherwise an ordinary character, only write if we have room,  }
            { else report that message was truncated.                       }
            if CurrentColumn<LineSize[StartLine+CurrentRow]-2 then
            begin
                Dest[DestIndex]:=Message[SourceIndex];
                Inc(DestIndex);
                Inc(CurrentColumn);
            end else PatcherStatus:=ptsMessageTruncated;
        end;

        { Get next character from message }
        Inc(SourceIndex);
    end;

    { If there are lines/characters in Dest still remaining, pad them off }
    while CurrentRow<NumLines do
    begin
        while CurrentColumn<LineSize[StartLine+CurrentRow]-2 do
        begin
            Dest[DestIndex]:=#32;
            Inc(DestIndex);
            Inc(CurrentColumn);
        end;
        Dest[DestIndex]:=#10;
        Dest[DestIndex+1]:=#0;
        Inc(DestIndex,2);
        CurrentColumn:=0;
        Inc(CurrentRow);
    end;

    { Report error if message is truncated }
    if SourceIndex<=Length(Message) then PatcherStatus:=ptsMessageTruncated;
end;

procedure PatchASCIIZ(Offset:Longint;MaxLength:Word;const Message:String);
begin
    if Length(Message)>MaxLength-1 then
    begin
        StrPCopy(PointerOffset(Offset),Copy(Message,1,MaxLength-1));
        PatcherStatus:=ptsMessageTruncated;
    end else begin
        StrPCopy(PointerOffset(Offset),Message);
        PatcherStatus:=ptsOk;
    end;
end;

procedure PatchFile(Offset:Longint;const FileName:String);
var
    InputFile:File;
    Length:Longint;
    InputBufSize:Word;
begin
    PatcherStatus:=ptsOk;
    FileMode:=0;
    Assign(InputFile,FileName);
    Reset(InputFile,1);
    if IOResult=0 then
    begin
        Length:=FileSize(InputFile);
        InputBufSize:=TransferBlockSize;
        while (Length>0) and (PatcherStatus=ptsOk) do
        begin
            if Length<TransferBlockSize then InputBufSize:=Length;
            BlockRead(InputFile,PointerOffset(Offset)^,InputBufSize);
            if IOResult=0 then
            begin
                Inc(Offset,InputBufSize);
                Dec(Length,InputBufSize);
            end else PatcherStatus:=ptsDiskError;
        end;
        Close(InputFile);
    end else PatcherStatus:=ptsFileNotFound;
end;

procedure PatchBytes(Offset:Longint;Length:Word;var Data:array of Byte);
begin
    Move(Data,PointerOffset(Offset)^,Length);
    PatcherStatus:=ptsOk;
end;

procedure WriteImage(Offset,Length:Longint;const FileName:String);
var
    OutputFile:File;
    OutputBufSize:Word;
begin
    PatcherStatus:=ptsOk;
    Assign(OutputFile,FileName);
    Rewrite(OutputFile,1);
    if IOResult=0 then
    begin
        OutputBufSize:=TransferBlockSize;
        while (Length>0) and (PatcherStatus=ptsOk) do
        begin
            if Length<TransferBlockSize then OutputBufSize:=Length;
            BlockWrite(OutputFile,PointerOffset(Offset)^,OutputBufSize);
            if IOResult=0 then
            begin
                Inc(Offset,OutputBufSize);
                Dec(Length,OutputBufSize);
            end else PatcherStatus:=ptsOutOfSpace;
        end;
        Close(OutputFile);
    end else PatcherStatus:=ptsFileCreateError;
end;

end.
